/* ##################################################################################
#
#  Python interface for Pixie's point cloud API.
#  Written as part of RIBKit (http://ribkit.sourceforge.net).
#  Copyleft (C) 2006 Konstantin Evdokimenko a.k.a Qew (qewerty@gmail.com)
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
################################################################################## */

#include <Python.h>
#include <ptcapi.h>

static PyObject *ErrorObject;

/* ----------------------------------------------------- */

/* Declarations for objects of type PointCloud */

typedef struct {
	PyObject_HEAD
	
	PtcPointCloud ptc;
	int nvars;
	char *types[256];
	char *names[256];
	float world2eye[16];
	float world2ndc[16];
	float format;
} PointCloudObject;

static PyTypeObject PointCloudType;

/* ---------------------------------------------------------------- */

static char PointCloud_get_points_number__doc__[] = 
"Returns number of points in PointCloud object."
;

static PyObject *
PointCloud_get_points_number(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	int npoints;
	PtcGetPointCloudInfo(self->ptc, "npoints", (void *)(& npoints));
		
	return Py_BuildValue("i", npoints);
}


static char PointCloud_get_bbox__doc__[] = 
"Returns bounding box of PointCloud object."
;

/* Not implemented yet */
static PyObject *
PointCloud_get_bbox(PointCloudObject *self, PyObject *args)
{
	PyErr_SetString(PyExc_NotImplementedError, 
                    "get_bbox");
    return NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;
	Py_INCREF(Py_None);
	return Py_None;
}


static char PointCloud_get_datasize__doc__[] = 
"Returns size of data in floats."
;

static PyObject *
PointCloud_get_datasize(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	int size;
	PtcGetPointCloudInfo(self->ptc, "datasize", (void *)(& size));
		
	return Py_BuildValue("i", size);
}


static char PointCloud_get_world2eye__doc__[] = 
"Returns world to eye transformation matrix."
;

/* Not implemented yet */
static PyObject *
PointCloud_get_world2eye(PointCloudObject *self, PyObject *args)
{
	PyErr_SetString(PyExc_NotImplementedError, 
                    "get_world2eye");
    return NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	float matrix[16];
	PtcGetPointCloudInfo(self->ptc, "world2eye", (void *)(& matrix));
		
	PyObject *value = Py_BuildValue("i", matrix);
	Py_XINCREF(value);
	return value;
}


static char PointCloud_get_world2ndc__doc__[] = 
"Return world to NDC transformation matrix."
;

/* Not implemented yet */
static PyObject *
PointCloud_get_world2ndc(PointCloudObject *self, PyObject *args)
{
	PyErr_SetString(PyExc_NotImplementedError, 
                    "get_world2ndc");
    return NULL;
                    
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
	
	float matrix[16];
	PtcGetPointCloudInfo(self->ptc, "world2ndc", (void *)(& matrix));
		
	return Py_BuildValue("i", matrix);
}


static char PointCloud_get_types__doc__[] = 
"Return tuple of types of PointCloud channels."
;

static PyObject *
PointCloud_get_types(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
	
	PyObject *tuple = PyTuple_New(self->nvars);
	if(!tuple)
		return NULL;
	int i;
	for(i = 0; i < self->nvars; i++)
	{
		PyTuple_SET_ITEM(tuple, i, PyString_FromString(self->types[i]));
	}
		
	Py_INCREF(tuple);
	return tuple;
}


static char PointCloud_get_names__doc__[] = 
"Return tuple of names of PointCloud channels."
;

static PyObject *
PointCloud_get_names(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	PyObject *tuple = PyTuple_New(self->nvars);
	if(!tuple)
		return NULL;
	int i;
	for(i = 0; i < self->nvars; i++)
	{
		PyTuple_SET_ITEM(tuple, i, PyString_FromString(self->names[i]));
	}
		
	Py_INCREF(tuple);
	return tuple;
}


static char PointCloud_finish__doc__[] = 
"Finish and close point cloud file."
;

static PyObject *
PointCloud_finish(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	PtcFinishPointCloudFile(self->ptc);
		
	Py_INCREF(Py_None);
	return Py_None;
}


static char PointCloud_close__doc__[] = 
"Close point cloud file."
;

static PyObject *
PointCloud_close(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	PtcClosePointCloudFile(self->ptc);
		
	Py_INCREF(Py_None);
	return Py_None;
}


static char PointCloud_read_data_point__doc__[] = 
""
;

/* WARINING! Bug with "PtcReadDataPoint(self->ptc, point, normal, &radius, data);"!
   Perhaps Pixie's bug. Or ...*/
static PyObject *
PointCloud_read_data_point(PointCloudObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
		
	float point[3];
	float normal[3];
	float radius;
	float *data;
	
	int datasize;
	PtcGetPointCloudInfo(self->ptc, "datasize", &datasize);
	data = (float *)malloc(sizeof(float)*datasize);
	
	PtcReadDataPoint(self->ptc, point, normal, &radius, data);
	
	printf("OK\n");
	
	PyObject *data_string =\
		PyString_FromStringAndSize((char *)data, datasize * sizeof(float));
	free(data);
	
	PyObject *point_tuple = PyTuple_New(3);
	if(!point_tuple)
		return NULL;
	PyTuple_SET_ITEM(point_tuple, 0, PyFloat_FromDouble((double)point[0]));
	PyTuple_SET_ITEM(point_tuple, 1, PyFloat_FromDouble((double)point[1]));
	PyTuple_SET_ITEM(point_tuple, 2, PyFloat_FromDouble((double)point[2]));
	
	PyObject *normal_tuple = PyTuple_New(3);
	if(!normal_tuple)
		return NULL;
	PyTuple_SET_ITEM(normal_tuple, 0, PyFloat_FromDouble((double)normal[0]));
	PyTuple_SET_ITEM(normal_tuple, 1, PyFloat_FromDouble((double)normal[1]));
	PyTuple_SET_ITEM(normal_tuple, 2, PyFloat_FromDouble((double)normal[2]));
	
	PyObject *radius_object = PyFloat_FromDouble((double)radius);
	
	
	PyObject *result_tuple = PyTuple_New(4);
	if(!result_tuple)
		return NULL;
	PyTuple_SET_ITEM(result_tuple, 0, point_tuple);
	PyTuple_SET_ITEM(result_tuple, 1, normal_tuple);
	PyTuple_SET_ITEM(result_tuple, 2, radius_object);
	PyTuple_SET_ITEM(result_tuple, 4, data_string);
		
	Py_INCREF(point_tuple);
	Py_INCREF(normal_tuple);
	Py_INCREF(radius_object);
	Py_INCREF(data_string);
	Py_INCREF(result_tuple);
	return result_tuple;
}


static char PointCloud_write_data_point__doc__[] = 
""
;

static PyObject *
PointCloud_write_data_point(PointCloudObject *self, PyObject *args)
{
	float point[3];
	float normal[3];
	float radius;
	float *data;
	PyObject *point_object;
	PyObject *normal_object;
	PyObject *data_object;
	
	if (!PyArg_ParseTuple(args, "OOfO", &point_object, &normal_object, &radius, &data_object))
		return NULL;
		
	if(!PyTuple_Check(point_object))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("argument 1 (point) must be a tuple of floats"));
		return NULL;
	}
	if(!PyTuple_Check(normal_object))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("argument 2 (normal) must be a tuple of floats"));
		return NULL;
	}
	if(!PyString_Check(data_object))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("argument 4 must be a string"));
		return NULL;
	}
	
	/* point */
	PyObject *ob = PyTuple_GetItem(point_object, 0);
	if(PyFloat_Check(ob))
	{
		point[0] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		point[0] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 0 of point's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	ob = PyTuple_GetItem(point_object, 1);
	if(PyFloat_Check(ob))
	{
		point[1] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		point[1] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 1 of point's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	ob = PyTuple_GetItem(point_object, 2);
	if(PyFloat_Check(ob))
	{
		point[2] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		point[2] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 2 of point's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	/* normal */
	ob = PyTuple_GetItem(normal_object, 0);
	if(PyFloat_Check(ob))
	{
		normal[0] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		normal[0] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 0 of normal's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	ob = PyTuple_GetItem(normal_object, 1);
	if(PyFloat_Check(ob))
	{
		normal[1] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		normal[1] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 1 of normal's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	ob = PyTuple_GetItem(normal_object, 2);
	if(PyFloat_Check(ob))
	{
		normal[2] = (float)PyFloat_AS_DOUBLE(ob);
	}
	else if(PyInt_Check(ob))
	{
		normal[2] = (float)PyInt_AS_LONG(ob);
	}
	else
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("item 2 of normal's tuple (argument 1) must be a float or int"));
		return NULL;
	}
	
	/* data */
	int buffersize = PyString_GET_SIZE(data_object);
	int datasize;
	PtcGetPointCloudInfo(self->ptc, "datasize", &datasize);
	if(buffersize != datasize * sizeof(float))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("size of data buffer must be equal to size of data of point cloud"));
		return NULL;
	}
	data = (float *)PyString_AsString(data_object);
	
	PtcWriteDataPoint(self->ptc, point, normal, radius, data);
		
	Py_INCREF(Py_None);
	return Py_None;
}


static struct PyMethodDef PointCloud_methods[] = {
 {"get_points_number",	(PyCFunction)PointCloud_get_points_number,	METH_VARARGS,	PointCloud_get_points_number__doc__},
 {"get_bbox",	(PyCFunction)PointCloud_get_bbox,	METH_VARARGS,	PointCloud_get_bbox__doc__},
 {"get_datasize",	(PyCFunction)PointCloud_get_datasize,	METH_VARARGS,	PointCloud_get_datasize__doc__},
 {"get_world2eye",	(PyCFunction)PointCloud_get_world2eye,	METH_VARARGS,	PointCloud_get_world2eye__doc__},
 {"get_world2ndc",	(PyCFunction)PointCloud_get_world2ndc,	METH_VARARGS,	PointCloud_get_world2ndc__doc__},
 {"get_types",	(PyCFunction)PointCloud_get_types,	METH_VARARGS,	PointCloud_get_types__doc__},
 {"get_names",	(PyCFunction)PointCloud_get_names,	METH_VARARGS,	PointCloud_get_names__doc__},
 {"finish",	(PyCFunction)PointCloud_finish,	METH_VARARGS,	PointCloud_finish__doc__},
 {"close",	(PyCFunction)PointCloud_close,	METH_VARARGS,	PointCloud_close__doc__},
 {"read_data",	(PyCFunction)PointCloud_read_data_point,	METH_VARARGS,	PointCloud_read_data_point__doc__},
 {"write_data",	(PyCFunction)PointCloud_write_data_point,	METH_VARARGS,	PointCloud_write_data_point__doc__},
 
	{NULL}		/* sentinel */
};

/* ---------- */


static PyObject *
PointCloud_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
	PointCloudObject *self;
	
	self = (PointCloudObject *)type->tp_alloc(type, 0);

	Py_INCREF(self);
	return (PyObject *)self;
}


static void
PointCloud_dealloc(PointCloudObject *self)
{
	PtcFinishPointCloudFile(self->ptc);
	
	int i;
	for(i = 0; i < self->nvars; i++)
	{
		free(self->types[i]);
		free(self->names[i]);
	}
	
	((PyObject *)(self))->ob_type->tp_free((PyObject *)(self));
}

static int
PointCloud_init(PointCloudObject *self, PyObject *args, PyObject *kwds)
{
	char *filename;
	
	PyObject *types;
	PyObject *names;
	if (!PyArg_ParseTuple(args, "sOO", &filename, &types, &names ))
	{
        return -1;
	}
	
	if(!PyTuple_Check(types))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("argument 2 must be a tuple of strings"));
		return -1;
	}
	if(!PyTuple_Check(names))
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("argument 3 must be a tuple of strings"));
		return -1;
	}
	
	int types_size = PyTuple_GET_SIZE(types);
	int names_size = PyTuple_GET_SIZE(names);
	if(types_size != names_size)
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("lengths of tuples (arguments 2 and 3) must be equal"));
		return -1;
	}
	if(types_size >= 256)
	{
		PyErr_SetObject(ErrorObject, PyString_FromString("lengths of tuples (arguments 2 and 3) must be less than 256"));
		return -1;
	}
		
		
	self->nvars = names_size;
	int i;
	for(i = 0; i < self->nvars; i++)
	{
		PyObject *item = PyTuple_GetItem(types, i);
		if(!PyString_Check(item))
		{
			PyErr_SetObject(ErrorObject, PyString_FromString("tuple of channel types (argument 2) must contain only strings"));
			return -1;
		}
		char *s = PyString_AsString(item);
		self->types[i] = (char *)malloc(sizeof(char)*(strlen(s)+1));
		if(!self->types[i])
		{
			PyErr_SetObject(ErrorObject, PyString_FromString("memory problem"));
			return -1;
		}
		strcpy(self->types[i], s);
	}
	for(i = 0; i < self->nvars; i++)
	{
		PyObject *item = PyTuple_GetItem(names, i);
		if(!PyString_Check(item))
		{
			PyErr_SetObject(ErrorObject, PyString_FromString("tuple of channel names (argument 3) must contain only strings"));
			return -1;
		}
		char *s = PyString_AsString(item);
		self->names[i] = (char *)malloc(sizeof(char)*(strlen(s)+1));
		if(!self->names[i])
		{
			PyErr_SetObject(ErrorObject, PyString_FromString("memory problem"));
			return -1;
		}
		strcpy(self->names[i], s);
	}
	
	for(i = 0; i < 16; i++)
		self->world2eye[i] = (i % 5) ? 0 : 1;
	
	self->ptc = PtcCreatePointCloudFile(filename,
	                                    self->nvars,
	                                    self->types,
	                                    self->names,
	                                    self->world2eye,
	                                    self->world2ndc,
	                                    &(self->format) );
	                                    
	return 0;
}

static char PointCloudType__doc__[] = 
"Pixie's point cloud."
;

static PyTypeObject PointCloudType = 
{
	PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "PointCloud",             /*tp_name*/
    sizeof(PointCloudObject),             /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)PointCloud_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    PointCloudType__doc__,           /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    PointCloud_methods,             /* tp_methods */
    0,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)PointCloud_init,      /* tp_init */
    0,                         /* tp_alloc */
    PointCloud_new,                 /* tp_new */
};

/* End of code for PointCloud objects */
/* -------------------------------------------------------- */

static PyObject *
ptc_open(PyObject *self, PyObject *args)
{
    char *filename;

    if (!PyArg_ParseTuple(args, "s", &filename))
        return NULL;
        
    PointCloudObject *file;
    
	file = (PointCloudObject *)PointCloud_new(& PointCloudType, NULL, NULL);
	
	file->ptc = PtcOpenPointCloudFile(filename, 
	                                  &(file->nvars),
	                                  file->types,
	                                  file->names);
    return (PyObject *)file;
}

/* List of methods defined in the module */

static struct PyMethodDef ptc_methods[] = {

	{"open",  ptc_open, METH_VARARGS,
     "Open point cloud file and return ptc.PointCloud object."},
     
	{NULL,	 (PyCFunction)NULL, 0, NULL}		/* sentinel */
};


/* Initialization function for the module (*must* be called initptc) */

static char ptc_module_documentation[] = 
"Python interface for Pixie's point cloud API.\n"
"Author: Konstantin Evdokimenko (qewerty@gmail.com).\n"
"License: GNU General Public License version 2.\n\n"
"Example:\n\n"
"import ptc\n"
"p = ptc.PointCloud('filename', ('color', 'point'), ('color1', 'point1'))\n"
"p.write_data((0, 0, 1.1), (1, 0, 0), 0.1, struct.pack('ffffff', 1., 0., 0., 4., 5., 6.))\n"
"p.finish()\n"
;

PyMODINIT_FUNC
initptc(void)
{
	PyObject *m, *d;
	
	if(PyType_Ready(&PointCloudType) < 0)
		return;

	/* Create the module and add the functions */
	m = Py_InitModule4("ptc", ptc_methods,
		ptc_module_documentation,
		(PyObject*)NULL,PYTHON_API_VERSION);

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);
	ErrorObject = PyErr_NewException("ptc.Error", NULL, NULL);
	PyDict_SetItemString(d, "Error", ErrorObject);
		
	PyModule_AddObject(m, "PointCloud", (PyObject *)&PointCloudType);
	
	/* Check for errors */
	if (PyErr_Occurred())
		Py_FatalError("can't initialize module ptc");
}

